########################################################
########################################################
## Purpose: estimate the model by MLE
##
## INPUTS:
## data: list()
##    $yM: T vector holding the macro observations
##    $yR: T x I x J array holding the transitions
##    $nR: T x I matrix holding the exposures
##    $zmat: T x k matrix of covariate observations
## init_par: vector with initial parameters
## nr_obs: only use the first nr_obs observations; if not provided (NULL), use all.
##
## OUTPUTS: list()
##    output of minimization nlm() using -average log-lik
##
## Author: Andre Lucas
## Version May 10, 2021
##
MLE <- function(data, init_par = NULL, nr_obs = NULL) {

  # initialize 
  if (GLOBAL_WITH_MACRO_COVARIATES) {
    if (is.null(data$zmatM)) nr_exogenousM <- 0 else nr_exogenousM <- ncol(data$zmatM)
  }
  if (GLOBAL_WITH_CREDIT_COVARIATES) {
    if (is.null(data$zmatR)) nr_exogenousR <- 0 else nr_exogenousR <- ncol(data$zmatR)
  }
  
  ## define local -log-lik function
  likelihood <- function(p){
    if ((max(abs(p)) > 10) | (sum(is.na(p)) > 0)) return(1e10)
    parameter <- get_parameters(p, nr_exogenousM = nr_exogenousM, nr_exogenousR = nr_exogenousR)
    output <- gasfilter(parameter, data, nr_obs = nr_obs)
    GLOBAL_PLOT_COUNTER <<- GLOBAL_PLOT_COUNTER + 1
    if (GLOBAL_PLOT_COUNTER >= GLOBAL_PLOT_FREQ) {
      plot_data(data, output)
      GLOBAL_PLOT_COUNTER <<- 0
    }
    return( -mean(output$likM.t + output$likR.t) )
  }
  likelihood.t <- function(p){
    parameter <- get_parameters(p, nr_exogenousM = nr_exogenousM, nr_exogenousR = nr_exogenousR)
    output <- gasfilter(parameter, data, nr_obs = nr_obs)
    return( -(output$likM.t + output$likR.t) )
  }

  ## initialize the optimization if no initial value provided
  if (is.null(init_par)) {
    init_par <- NULL
    if (GLOBAL_WITH_MACRO_FRAILTY) init_par <- c(init_par, log(1.269))
    if (GLOBAL_WITH_CREDIT_FRAILTY) init_par <- c(init_par, log(0.2))
    if (GLOBAL_WITH_MACRO_FRAILTY & !GLOBAL_EWMA_MACRO) init_par <- c(init_par, log(0.964/0.036))
    if (GLOBAL_WITH_CREDIT_FRAILTY & !GLOBAL_EWMA_CREDIT) init_par <- c(init_par, log(0.935/0.065))
    init_par <- c(init_par, 
                  log(1.337), # s2 log
                  mean(data$yM), # muM
                  -6.554,  6.014, log(4.8),  log(12.3), # muR
                  NULL)
    if (GLOBAL_WITH_MACRO_FRAILTY_IN_CREDITS) init_par <- c(init_par, 0, 0)
    if (GLOBAL_WITH_CREDIT_FRAILTY) init_par <- c(init_par, 0)
    if (GLOBAL_WITH_MACRO_COVARIATES) init_par <- c(init_par, rep(0, ncol(data$zmatM)))
    if (GLOBAL_WITH_CREDIT_COVARIATES) init_par <- c(init_par, rep(0, ncol(data$zmatR) * dim(data$yR)[2]))
  }

  ## perform the mle and return the result
  my_mle <- nlm(likelihood,init_par, print.level = 2, hessian = TRUE, 
                iterlim = 500, steptol = 1e-6)
  my_mle$output <- gasfilter(get_parameters(my_mle$estimate, nr_exogenousM = nr_exogenousM, nr_exogenousR = nr_exogenousR), data)
  
  ## compute jacobian and hessian
  print("## Computing the OPG")
  OPG <- jacobian(likelihood.t, my_mle$estimate, method = "simple")
  iT <- nrow(OPG)
  my_mle$loglik <- -iT * my_mle$minimum
  OPG <- t(OPG) %*% OPG / iT
  H <- my_mle$hessian
  Hinv <- solve((H + t(H)) / 2)
  ## store three types of covariance matrices
  my_mle$sandwich <- Hinv %*% OPG %*% Hinv / iT
  my_mle$hessian_inv <- Hinv / iT
  my_mle$opg_inv <- solve(OPG) / iT
  # store the transformed parameters (of interest) and their covariance matrices
  print("## Computing the Jacobian and untransformed parameters of interest")
  org_parameterize <- function(x) {
    return(
      get_parameters(
        get_parameters(x, nr_exogenousM = nr_exogenousM, nr_exogenousR = nr_exogenousR), 
        nr_exogenousM = nr_exogenousM, nr_exogenousR = nr_exogenousR,
        setpar = TRUE, usetransform = FALSE
      )
    )
  }
  J <- jacobian(org_parameterize, my_mle$estimate)
  my_mle$Jacobian_trafo <- J
  my_mle$t_estimate <- org_parameterize(my_mle$estimate)
  my_mle$t_sandwich <- J %*% my_mle$sandwich %*% t(J)
  my_mle$t_hessian_inv <- J %*% my_mle$hessian_inv %*% t(J)
  my_mle$t_opg_inv <- J %*% my_mle$opg_inv %*% t(J)
  
  # return value
  return(my_mle)
}
