## Libraries
library(openxlsx)
library(lubridate)
library(stringi)
library(zoo)
library(numDeriv)

###############################################################################
###############################################################################
## Different models
##
## Model 1: Constant
## Model 2: Macro frailty
## Model 3: Macro frailty in credits
## Model 4: Macro frailty in credits & credit frailty
## Model 5: Macro frailty with credit covariate
## Model 6: Macro frailty, credit frailty & credit covariate
## 
## Model 31: Model 3 + subsidy
## Model 41: Model 4 + subsidy
## Model 51: Model 5 + subsidy
## Model 61: Model 6 + subsidy
##
## Variable subsidy is c('SubsIP','TarpIP','FedSurplusIP') or a subset thereof 
##
## Binary: precovid  (if true, use precovid data, otherwise complete data set)
##
##
## 1 jan 2000 ->
###############################################################################
###############################################################################
models <- c(5,3,4,6, 31,41,51,61)
models <- c(31,41,51,61)
models <- c(4)

## Estimate pre-covid (1), full sample (2), or try both c(1,2)
# optionset <- c(1,2)
optionset <- c(1)

## Set subsidy variable combinations to try out, possibilities in
## c('SubsIP','TarpIP','FedSurplusIP')
subsidy_sets <- list( 
  s1 = c('SubsIP')
  # s1 = c('TarpIP'),
  # s1 = c('FedSurplusIP'),
  # s1 = c('SubsIP', 'TarpIP'),
  # s1 = c('SubsIP', 'FedSurplusIP'),
  # s1 = c('SubsIP', 'TarpIP', 'FedSurplusIP'),
  # s1 = c('SubsIP', 'TarpIP', 'FedSurplusIP','AigIP')
)
# subsidy_sets <- list( 
#   s1 = c('AigIP')
# )
# If subsidy = 'SubsIP': current paper setup
# To consider only models with subsidy, change model to c(31,41,51,61)

## lag macro (as covariate) by a month (1) or do not (0) lag as covariate ... 
## macro as dependent is times the same as rating experience
macro_lag <- 1

## set folder to fetch data
datadirectory <- ""
## rating file name
rating_filename <- "transitions_2scale_RatingDate.csv"
## set start and end date of sample to be used
start_date <- as.Date("1998-01-01")
end_date <- as.Date("2021-09-30")
## set start of covid sample
covid_start_date <- as.Date("2020-01-01")
## output file name for all estimates across models, samples, controls
timing_name <- stri_replace_all(date(), ".", regex = "[:;/-]")
Total_results_name <- paste0("Total-results ", timing_name, ".csv")
Output_directory_name <- paste0("./output ", timing_name, "/")
if (!dir.exists(Output_directory_name)) dir.create(Output_directory_name)


## read packages
for (engine in c("gasfilter", "get_parameters", "read_data", "plot_data",
                 "MLE", "get_paper_parameters", "forecast_credit")) source(paste(engine, ".R", sep = ""))
options(digits = 8)




################################# FUNCTIONS ##################################
################################# FUNCTIONS ##################################
################################# FUNCTIONS ##################################
################################# FUNCTIONS ##################################
################################# FUNCTIONS ##################################
################################# FUNCTIONS ##################################
read_transform_data <- function(filename) {
  mydata <- read.csv(filename)
  mycolumns <- c("IG.IG", "IG.NP", "IG.D", "NP.IG", "NP.NP", "NP.D")
  return( mydata[ , c("Mth.Year", mycolumns)] )
}

read_data <- function(model,precovid=FALSE,subsidy){
  
  ## Ratings data
  ratings.df <- read_transform_data(paste0(datadirectory, rating_filename))
  ratings.df$Mth.Year <- as.Date(ratings.df$Mth.Year)
  ratings.df <- subset(ratings.df, (Mth.Year >= start_date) & (Mth.Year <= end_date))
  yRm <- data.matrix(ratings.df[ , 2:7])
  if(precovid){
    iT <- nrow(yRm) - sum(0 + (ratings.df$Mth.Year >= covid_start_date))
  } else {
    iT <- nrow(yRm)
  }
  yR <- array(NA, dim = c(iT, 2, 3))
  for (i1 in sequence(iT)){
    yR[i1, , ] <- matrix(yRm[i1, ], nrow = 2, ncol = 3, byrow = TRUE)
  } 
  
  ## read macro dataframe
  macro.df <- read.csv(paste(datadirectory, "macroJBFpreprocessed.csv", sep = ""))
  macro.df$observation_date <- as.Date(macro.df$observation_date)
  ## transform TarpIP from percentage into basis points
  macro.df$TarpIP <- macro.df$TarpIP * 100
  # lag macro by a month macro_lag days ... 
  # do *not* use subset observation_date <= end_date - macro_lag*31, as the
  # length of the series is automatically catered to the length of the rating 
  # series 2 lines below (using iT)
  macro.df <- subset(macro.df, macro.df$observation_date >= start_date - macro_lag * 31)
                       
  ## Macro data
  yM <- macro.df[sequence(iT) + macro_lag, "IPgrowth"]
  zmatR <- matrix(0,nrow=iT)
  colnames(zmatR) <- "None"
  
  ## Subsidy data
  if (model == 5 || model == 6){
    zmatR <- as.matrix(yM, nrow = iT)
    colnames(zmatR) <- c("IPgrowth")
  } else if (model == 31 || model == 41){
    zmatR <- as.matrix(macro.df[sequence(iT), subsidy])
    colnames(zmatR) <- c(subsidy)
  } else if (model == 51 || model == 61){
    zmatR <- as.matrix(cbind(yM, macro.df[sequence(iT), subsidy]))
    colnames(zmatR) <- c("IPgrowth", subsidy)
  }
  
  return(list(yM = yM, yR = yR, zmatR = zmatR, 
              date = ratings.df$Mth.Year[sequence(iT)])) 
}

compute_stderr <- function(mle_res){
  stderr <- mle_res
  stderr.sand <- sqrt(diag(mle_res$t_sandwich))
  stderr$sand <- data.frame(stderr.sand)
  stderr.hess <- sqrt(diag(mle_res$t_hessian_inv))
  stderr$hess <- data.frame(stderr.hess)
  stderr.opg  <- sqrt(diag(mle_res$t_opg_inv))
  stderr$opg <- data.frame(stderr.opg)
  
  return(stderr)
}

fit_mle <- function(data,mle_res,filter_res){
  lik <- list()
  k <- length(mle_res$t_estimate)
  likM <- sum(filter_res$likM.t)
  lik$z.lik.M <- likM
  # lik$AIC <- 2*k - 2*likM
  # lik$BIC <- k*log(NROW(data$yM)) - 2*likM
  likR <- sum(filter_res$likR.t)
  lik$z.lik.R <- likR
  lik$z.AIC.R <- 2*k - 2*likR
  lik$z.BIC.R <- k*log(NROW(data$yM)) - 2*likR
  # lik$tot  <- likM + likR
  # lik$tot2 <- mle_res$loglik
  
  return(lik)
}

set_globals <- function(model){
  ## generic sets 
  GLOBAL_PLOT_COUNTER <<- 1
  GLOBAL_PLOT_FREQ <<- 20
  GLOBAL_USE_ANNUAL_IP_GROWTH <<- TRUE
  GLOBAL_EWMA_MACRO <<- FALSE
  GLOBAL_EWMA_CREDIT <<- FALSE
  GLOBAL_WITH_MACRO_FRAILTY <<- TRUE
  GLOBAL_WITH_MACRO_COVARIATES <<- FALSE
  
  if(model == 1){
    # ## set the MACRO part of the model
    GLOBAL_WITH_MACRO_FRAILTY <<- FALSE
    # ## set the CREDIT part of the model
    GLOBAL_WITH_CREDIT_FRAILTY <<- FALSE
    GLOBAL_WITH_MACRO_FRAILTY_IN_CREDITS <<- FALSE
    GLOBAL_WITH_CREDIT_COVARIATES <<- FALSE
  }
  else if (model == 2){
    # ## set the CREDIT part of the model
    GLOBAL_WITH_CREDIT_FRAILTY <<- FALSE
    GLOBAL_WITH_MACRO_FRAILTY_IN_CREDITS <<- FALSE
    GLOBAL_WITH_CREDIT_COVARIATES <<- FALSE
  }
  else if (model == 3 || model == 31){
    # ## set the CREDIT part of the model
    GLOBAL_WITH_CREDIT_FRAILTY <<- FALSE
    GLOBAL_WITH_MACRO_FRAILTY_IN_CREDITS <<- TRUE
    GLOBAL_WITH_CREDIT_COVARIATES <<- (model == 31)
  }
  else if (model == 4 || model == 41){
    # ## set the CREDIT part of the model
    GLOBAL_WITH_CREDIT_FRAILTY <<- TRUE
    GLOBAL_WITH_MACRO_FRAILTY_IN_CREDITS <<- TRUE
    GLOBAL_WITH_CREDIT_COVARIATES <<- (model == 41)
  }
  else if (model == 5 || model == 51){
    # ## set the CREDIT part of the model
    GLOBAL_WITH_CREDIT_FRAILTY <<- FALSE
    GLOBAL_WITH_MACRO_FRAILTY_IN_CREDITS <<- FALSE
    GLOBAL_WITH_CREDIT_COVARIATES <<- TRUE
  }
  else if (model == 6 || model == 61){
    # ## set the CREDIT part of the model
    GLOBAL_WITH_CREDIT_FRAILTY <<- TRUE
    GLOBAL_WITH_MACRO_FRAILTY_IN_CREDITS <<- FALSE
    GLOBAL_WITH_CREDIT_COVARIATES <<- TRUE
  }
}


################################# MAIN ##################################
################################# MAIN ##################################
################################# MAIN ##################################
################################# MAIN ##################################
################################# MAIN ##################################
parameter.df <- data.frame()
for (subsidy in subsidy_sets) for (myoptions in optionset) {
  
  if(myoptions==1){
    precovid=TRUE
  } else {
    precovid=FALSE
  }

  ## loop over models, but skip re-estimating models without subsidy
  ## variables unnecessarily (models 1..6)
  for(mod in models) if ((mod > 10) | 
                         ((mod <=10) & prod(subsidy %in% subsidy_sets[[1]]) &
                          prod(subsidy_sets[[1]] %in% subsidy))
  ) {
    my_data <- read_data(model=mod,precovid=precovid,subsidy=subsidy)
    nr_exogenousM <- ifelse(is.null(my_data$zmatM), 0, ncol(my_data$zmatM))
    nr_exogenousR <- ifelse(is.null(my_data$zmatR), 0, ncol(my_data$zmatR))
    set_globals(mod)
    
    ## estimate the model
    mle_out <- MLE(my_data)
    theta_hat <- get_parameters(mle_out$estimate, nr_exogenousM = nr_exogenousM, 
                              nr_exogenousR = nr_exogenousR)
    filter_out <- gasfilter(theta_hat, my_data)
    # plot_data(my_data, filter_out)
  
    ## Parameters
    # parameters <- data.frame(theta_hat)
    parameters <- mle_out$t_estimate

    ## Standard errors
    sderr <- compute_stderr(mle_out)
    sderr <- data.frame(sderr$sand, sderr$hess, sderr$opg)

    ## Fit
    mle_out$fit <- unlist(fit_mle(data=my_data,mle_res=mle_out,filter_out))

    ## Prepare output file
    mod <- toString(mod)
    subdat <- paste0(stri_sub(subsidy, to = 1), collapse = "")
    if(!precovid){
      sam <- 'full'
    } else{
      sam <- 'precov'
    }
  
    # path <- 'C:/Users/jtg540/Dropbox/Dubinova-Lucas-Telg/new results/'
    # path <- "../../new results/"
    path <- ""
    filename <- paste0(path,"Results Model ",mod," ", sam, " (",subdat,")",".xlsx", sep = "")

    # give unique names to parameter estimates for storage
    # across models
    model_name <- paste(mod, sam, subdat, sep = ".")
    theta_hat <- get_parameters(theta_hat, nr_exogenousM = nr_exogenousM, 
                                nr_exogenousR = nr_exogenousR, 
                                setpar = TRUE, usetransform = FALSE, 
                                exog_names = colnames(my_data$zmatR))
    parameter.df[ names(theta_hat), model_name] <- theta_hat
    parameter.df[ names(mle_out$fit), model_name] <- c(as.matrix(mle_out$fit))
    theta_hat_s <- get_parameters(sderr[ , 1], nr_exogenousM = nr_exogenousM, 
                                nr_exogenousR = nr_exogenousR, usetransform = FALSE) 
    theta_hat_s <- get_parameters(theta_hat_s, nr_exogenousM = nr_exogenousM, 
                                nr_exogenousR = nr_exogenousR, 
                                setpar = TRUE, usetransform = FALSE, 
                                exog_names = colnames(my_data$zmatR))
    parameter.df[ names(theta_hat), paste(model_name, "s", sep = ".")] <- theta_hat_s
    write.csv(parameter.df, file = paste0(Output_directory_name, Total_results_name))
    saveRDS(mle_out, file = paste0(Output_directory_name, "mle_out", model_name, ".Rdat"))
    saveRDS(filter_out, file = paste0(Output_directory_name, "filter", model_name, ".Rdat"))

    
    ## Write to Excel
    # results <- list("Parameters" = parameters, "Standard errors" = sderr, "Likelihood" = fit)
    # openxlsx::write.xlsx(results, file =  filename)
  }
}
