require("zoo")


forecast_credit <- function(parameter_vector, data, which_times = NULL, horizon = 1,
                            scenarios = 1, portfolio = NULL, 
                            fquantiles = c(0.99, 0.95, 0.9, 0.75, 0.5, 0.25, 0.1, 0.05, 0.01),
                            return_all_sim = FALSE, print_life_sign = 0,
                            stress_scenario = NULL) {

  # initialize
  nr_in_ratings <- dim(data$yR)[2]
  nr_out_ratings <- dim(data$yR)[3]
  nr_exogenousM <- ifelse(is.null(data$zmatM), 0, ncol(data$zmatM))
  nr_exogenousR <- ifelse(is.null(data$zmatR), 0, ncol(data$zmatR))
  horizon <- ifelse(as.integer(horizon) <= 0, 1, as.integer(horizon))
  
  # check forecast times; if NULL, forecast last observation from time T - horizon,
  # where T = dim(data$yR)[1], and data$yR is an array of T x I x J
  if (is.null(which_times)) which_times <- c(dim(data$yR)[1])
  if (is.null(stress_scenario)) stress_scenario <- rep(0, horizon)
  if (length(stress_scenario) != horizon) {print("Stress scenario wrong length compared to horizon"); stop(0)}

  # get the parameters
  parameter <- get_parameters(parameter_vector, nr_in_ratings, nr_out_ratings, nr_exogenousM, nr_exogenousR)
  
  # initialize output and index
  if (return_all_sim == TRUE) {
    all_sims_cum <- array(NA, dim = c(length(which_times), scenarios, nr_in_ratings, nr_out_ratings))
    all_sims_marg <- array(NA, dim = c(length(which_times), horizon, scenarios, nr_in_ratings, nr_out_ratings))
    all_sims_macro <- array(NA, dim = c(length(which_times), scenarios))
  }
  forecast_bands_cum <- array(NA, dim = c(length(which_times), length(fquantiles), nr_in_ratings, nr_out_ratings))
  forecast_bands_cum_p <- forecast_bands_cum
  forecast_bands_marg <- forecast_bands_cum
  forecast_bands_marg_p <- forecast_bands_cum
  forecast_macro <- array(NA, dim = c(length(which_times), length(fquantiles)))
  realization_cum <- array(NA, dim = c(length(which_times), nr_in_ratings, nr_out_ratings))
  realization_marg <- realization_cum
  idx <- 1
  
  for (forecast_time in which_times) {
    
    # print lifesign
    if (as.integer(print_life_sign) > 0) print(paste("... forecasting time point", forecast_time))
    
    # run the filter and construct the scenario paths
    output <- gasfilter(parameter, data, horizon = horizon, nr_obs = forecast_time - horizon, 
                        scenarios = scenarios, loanportfolio = portfolio, print_life_sign,
                        stress_scenario = stress_scenario)
    
    # cumulative forecast over horizon, or end of horizon forecast
    {
      # return all simulated empirical default freqs if asked for
      if (return_all_sim == TRUE) {
        all_sims_cum[idx, , , ] <- output$forecasts$pi_emp_cum[ , 1:nr_in_ratings, ]
        all_sims_macro[idx, ] <- output$forecasts$yM[ , horizon]
      }
      
      # store the macro forecast
      forecast_macro[idx, ] <- quantile(output$forecasts$yM[ , horizon], fquantiles)
        
      # obtain the forecast quantiles
      for (in_rating in sequence(nr_in_ratings)) {
        for (out_rating in sequence(nr_out_ratings)) {
          forecast_bands_cum[idx, , in_rating, out_rating] <- quantile(output$forecasts$pi_emp_cum[ , in_rating, out_rating], fquantiles)
          forecast_bands_cum_p[idx, , in_rating, out_rating] <- quantile(output$forecasts$pi_cum[ , in_rating, out_rating], fquantiles)
        }
      }
      
      # compute the realization
      pi_matrix <- diag(nr_out_ratings)
      pi_cum <- diag(nr_out_ratings)
      for (i1 in 1:horizon) {
        nr_exposures <- rowSums(data$yR[forecast_time - horizon + i1, , ])
        pi_matrix[1:nr_in_ratings, ] <- data$yR[forecast_time - horizon + i1, , ] / matrix( rep(nr_exposures, nr_out_ratings), nrow = nr_in_ratings)
        pi_cum <- pi_cum %*% pi_matrix
      }
      realization_cum[idx, , ] <- pi_cum[1:nr_in_ratings, ]

    } 
    # marginal forecast over horizon, or end of horizon forecast
    {

      # return all simulated empirical default freqs if asked for
      if (return_all_sim == TRUE) all_sims_marg[idx, , , , ] <- output$forecasts$yR
      
      # obtain the forecast quantiles
      for (in_rating in sequence(nr_in_ratings)) {
        for (out_rating in sequence(nr_out_ratings)) {
          forecast_bands_marg[idx, , in_rating, out_rating] <- quantile(output$forecasts$yR[ , horizon, in_rating, out_rating] / rowSums(output$forecasts$yR[ , horizon, in_rating, ]), fquantiles)
        }
      }
      
      # compute the realization
      for (in_rating in sequence(nr_in_ratings)) {
        realization_marg[idx, in_rating, ] <- data$yR[forecast_time, in_rating, ] / sum(data$yR[forecast_time, in_rating, ])
      }

      # duplicate to have same outputs as cumulative == TRUE  
      forecast_bands_marg_p[idx, , , ] <- forecast_bands_marg[idx, , , ]
    }
    
    # update the index
    idx <- idx + 1
    
  }
  
  # return the bands
  if (return_all_sim == TRUE) {
    return(list(forecast_bands_marg = forecast_bands_marg, forecast_bands_marg_p = forecast_bands_marg_p, realization_marg = realization_marg, all_sims_marg = all_sims_marg,
                forecast_bands_cum = forecast_bands_cum, forecast_bands_cum_p = forecast_bands_cum_p, realization_cum = realization_cum, all_sims_cum = all_sims_cum,
                forecast_macro = forecast_macro, all_sims_macro = all_sims_macro))
  } else {
    return(list(forecast_bands_marg = forecast_bands_marg, forecast_bands_marg_p = forecast_bands_marg_p, realization_marg = realization_marg, 
                forecast_bands_cum = forecast_bands_cum, forecast_bands_cum_p = forecast_bands_cum_p, realization_cum = realization_cum,
                forecast_macro = forecast_macro))
  }
}

