########################################################
########################################################
## Purpose: run the gas filter for the mixed normal-binomial
##    model, possibly with post sample simulations for forecasting
##
## INPUTS:
## parameter: list()
##    $muM: intercept of macro
##    $omegaM: K x 1 matrix with coefficients of covariates in macro equation
##    $s2: variance of error term
##    $muR: I x (J-1) matrix of intercepts for the ordered logit,
##          with I nr of input ratings, and J number of output 
##          rating categories
##    $gammaR: I x 1 matrix of coefficients for macro factor
##    $deltaR: I x 1 matrix of coefficients for frailty factor
##    $omegaR: K x 1 matrix with coefficients of covariates in ratings equation
## data: list()
##    $yM: T vector holding the macro observations
##    $yR: T x I x J array holding the transitions
##    $nR: T x I matrix holding the exposures
##    $zmat: T x K matrix of covariate observations
## horizon: integer (default = 0); if positive, then at the end of the sample,
##    simulate paths forward of length horizon of the state, the observables, 
##    and the probabilities for forecasting purposes. Forecasting is started from
##    the end of the sample. If covariates are part of the model, these need to
##    be there for the forecasting period, otherwise an error stops the function.
## stress_scenario: a vector of length horizon added (as additive outlier) to the
##    simulations of the macro realization over the forecasting period
## nr_obs: integer, giving the index of the last observation to be used. The 
##    forecasts start at nr_obs+1. If NULL, nr_obs is the last observation in yR
## scenarios: integer, the number of paths to be simulated for the forecasts.
## loanportfolio: initial portfolio of loans at start of forecast horizon; if 
##    NULL then use last exposures after the sample
##
## OUTPUTS: dataframe with elements
##    $lik.t: the log likelihood contributions
##    $fM: the estimated macro factor path
##    $fR: the estimated credit factor path
##    $IG_IG: transition probability from investment grade to investment grade
##    $IG_NP: transition probability from investment grade to non-prime
##    $IG_D: transition probability from investment grade to default
##    $NP_IG: transition probability from non-prime to investment grade
##    $NP_NP: transition probability from non-prime to non-prime
##    $NP_D: transition probability from non-prime to default
## if integer(horizon) > 0, then a list(), with first element $output as above,
##    $forecasts: list() with
##        $fM: scenarios x horizon with simulated macro factors
##        $yM: scenarios x horizon with simulated macros
##        $fR: scenarios x horizon with simulated credit factors
##        $yR: scenarios x horizon x I x J with simulated transition counts
##        $pi: scenarios x horizon x I x J with simulated probabilities
##        $pi_cum: scenarios x J x J with simulated product integrals, assuming that
##             states I+1...J are absorbing
##
## Author: 
## version May 8, 2021: Andre Lucas
## update: May 26, 2021: Andre Lucas
##
gasfilter <- function(parameter, data, horizon = 0, 
                      nr_obs = NULL, scenarios = 1, 
                      loanportfolio = NULL, print_life_sign = 0,
                      stress_scenario = NULL) {
  
  # initialize settings
  if (is.null(nr_obs)) nr_obs <- dim(data$yR)[1]
  nr_in_ratings <- nrow(parameter$muR)
  nr_out_ratings <- ncol(parameter$muR) + 1
  
  # check arguments
  horizon <- max(0, as.integer(horizon))
  scenarios <- max(1, as.integer(scenarios))
  if (is.null(loanportfolio)) {
    # take the exposures after the last period transitions
    loanportfolio <- colSums(data$yR[nr_obs, , sequence(nr_out_ratings - 1)])
  } else if (length(loanportfolio) == 1) {
    loanportfolio <- rep(loanportfolio, nr_in_ratings) 
  } else if (length(loanportfolio) != nr_in_ratings) {
    stop(paste("Error filter simulation: length loanportfolio", length(loanportfolio), 
               "not equal to nr rating classes", nr_in_ratings))
  }

  # check if exogenous variables present
  if (!GLOBAL_WITH_MACRO_COVARIATES) {
    zmatM <- matrix(0, nrow = nr_obs + horizon, ncol = 1)
  } else {
    zmatM <- data$zmatM %*% parameter$omegaM
  }
  if (!GLOBAL_WITH_CREDIT_COVARIATES) {
    zmatR <- matrix(0, nrow = nr_obs + horizon, ncol = nr_in_ratings)
  } else {
    zmatR <- matrix(data$zmatR %*% parameter$omegaR, nrow = nr_obs)
  }
  if ((nrow(zmatM) < nr_obs+ horizon) | (nrow(zmatR) < nr_obs+ horizon)) {
    stop(paste("Filter error: cannot forcast", horizon, "periods with unavailable future covariates"))
  }
  
  # initialize the bivariate filter
  f.t <- c(0,0)
  macro_init <- 5; f.t[1] <- mean(data$yM[1:macro_init] - zmatM[1:macro_init])
  output <- data.frame(
    "fM" = rep(NA, nr_obs + horizon), "fR" = NA, likM.t = NA, likR.t = NA,
    zmatM = zmatM, zmatR = zmatR)
  
  # initialize the forecasting storage
  if (horizon > 0) {
    forecasts <- list()
    forecasts$fM <- array(NA, dim = c(scenarios, horizon))
    forecasts$yM <- array(NA, dim = c(scenarios, horizon))
    forecasts$fR <- array(NA, dim = c(scenarios, horizon))
    forecasts$yR <- array(NA, dim = c(scenarios, horizon, nr_in_ratings, nr_out_ratings))
    forecasts$pi <- array(NA, dim = c(scenarios, horizon, nr_in_ratings, nr_out_ratings))
    forecasts$pi_cum <- array(NA, dim = c(scenarios, nr_out_ratings, nr_out_ratings))
    forecasts$pi_emp_cum <- array(NA, dim = c(scenarios, nr_out_ratings, nr_out_ratings))
  }
  pi_tmp <- diag(nr_out_ratings)
  
  # compute fixed scaling matrices
  S.t.sub <- array(0, dim = c(nr_in_ratings,2,2))
  for (i1 in sequence(nr_in_ratings)) {
    S.t.sub[i1, , ] <- matrix(c(parameter$gammaR[i1]^2, parameter$gammaR[i1] * parameter$deltaR[i1],
                                parameter$gammaR[i1] * parameter$deltaR[i1], parameter$deltaR[i1]^2), 
                              ncol = 2, nrow = 2)
  }
  
  # define the local function for one step in the filter
  # INPUTS:
  # parameter: a list with the parameters (see comment of outer function)
  # f.t: vector with the states
  # yM: scalar macro data
  # yR: matrix I x J with transition counts
  # zmatM: covariates times coefficients in macro equation
  # zmatR: covariates times coefficients in credit equation
  #
  # OUTPUTS: a list with
  #     $f.t: the NEXT state after the GAS update
  #     $likM.t: likelihood contribution of macro part
  #     $likR.t: likelihood contribution of credit part
  #     $pi_matrix: I x J matrix with fitted probabilities
  compute_pi_matrix <- function(parameter, f.t, zmatM, zmatR) {
    # compute the probabilities as the pi matrix
    pitilde_matrix <- parameter$muR + ( matrix(zmatR) + parameter$gammaR * f.t[1] + parameter$deltaR * f.t[2]) %*% t(c(1,1))
    pitilde_matrix <- cbind(1, 1 / (1 + exp(-pitilde_matrix)), 0)
    pi_matrix <- pitilde_matrix[ , sequence(nr_out_ratings)] -
      pitilde_matrix[ , 1 + sequence(nr_out_ratings)]
    pidot_matrix <- 
      pitilde_matrix[ , sequence(nr_out_ratings)] * ( 1 - pitilde_matrix[ , sequence(nr_out_ratings)] ) -
      pitilde_matrix[ , 1 + sequence(nr_out_ratings)] * ( 1 - pitilde_matrix[ , 1 + sequence(nr_out_ratings)] )
    return(list(pi_matrix = pi_matrix, pidot_matrix = pidot_matrix))
  }
  filter_one_step <- function(parameter, f.t, yM, yR, zmatM, zmatR) {
    # compute the probabilities as the pi matrix
    tmp <- compute_pi_matrix(parameter, f.t, zmatM, zmatR)
    pi_matrix <- tmp$pi_matrix
    pidot_matrix <- tmp$pidot_matrix

    # compute the likelihood contribution
    eM <- yM - f.t[1] - zmatM
    likM.t <- -0.5 * log( 2 * pi * parameter$s2 ) -0.5 * eM^2 / parameter$s2
    likR.t <- sum( yR * log(pi_matrix) )
    
    # compute the score
    sR.i <- rowSums(yR * pidot_matrix / pi_matrix)
    nabla <- c(
      eM / parameter$s2 + sR.i %*% parameter$gammaR,
      sR.i %*% parameter$deltaR
    ) 
    
    # compute the scaling
    c.i <- rowSums( pidot_matrix * pidot_matrix / pi_matrix ) * rowSums(yR)
    S.t <- 0
    for (i2 in sequence(nr_in_ratings)) S.t <- S.t + c.i[i2] * S.t.sub[i2, , ]
    S.t[1,1] <- S.t[1,1] + 1/parameter$s2
    if (GLOBAL_WITH_MACRO_FRAILTY & GLOBAL_WITH_CREDIT_FRAILTY) {
      S.t <- solve(S.t) # + 0.1 * (det(S.t) < 1e-4))
    } else {
      aid2 <- matrix(0, ncol = 2, nrow = 2)
      if (GLOBAL_WITH_MACRO_FRAILTY) aid2[1,1] <- 1 / S.t[1,1]
      if(GLOBAL_WITH_CREDIT_FRAILTY) aid2[2,2] <- 1 / S.t[2,2]
      S.t <- aid2
    }
    
    # update f.t
    f.t <- (1 - parameter$beta) * c(parameter$muM, 0) + parameter$beta * f.t + parameter$alpha * (S.t %*% nabla)
    
    # return
    return(list(f.t = f.t, likM.t = likM.t, likR.t = likR.t, pi_matrix = pi_matrix))
  }
  
  # define the local simulation routine for this model
  simulate_gas <- function(parameter, f.t, pi_matrix, portfolio) {
    # check parameter: break on omega present and not zero
    if (!is.null(parameter$omegaM)) if (max(abs(c(parameter$omegaM))) > 0) {print("Cannot simulate with exogenousM present"); stop(0)}
    if (!is.null(parameter$omegaR)) if (max(abs(c(parameter$omegaR))) > 0) {print("Cannot simulate with exogenousR present"); stop(0)}
    yM <- f.t[1] + sqrt(parameter$s2) * rnorm(1)
    yR <- matrix(0, nrow = nr_in_ratings, ncol = nr_out_ratings)
    for (i1 in sequence(nr_in_ratings)) yR[i1, ] <- rmultinom(1, portfolio[i1], pi_matrix[i1, ])
    return(list(yM = yM, yR = yR))
  }
  
  # loop over all observations plus the forecast horizon
  for (i1 in sequence(nr_obs + min(horizon, 1))) {
    
    # if within sample, update filter, otherwise enter forecasting mode
    if (i1 <= nr_obs) {
      # store the time varying factor
      output[i1, c("fM", "fR")] <- f.t 
      # make a step in the filter    
      tmp <- filter_one_step(parameter, f.t, data$yM[i1], data$yR[i1, , ], zmatM[i1, ], zmatR[i1, ])
      f.t <- tmp$f.t
      # store the outputs
      output[i1, c("IG_IG", "NP_IG", "IG_NP", "NP_NP", "IG_D", "NP_D")] <- c(tmp$pi_matrix)
      output$likM.t[i1] <- tmp$likM.t
      output$likR.t[i1] <- tmp$likR.t
    } else {
      # store the initial f.t
      f.t.original <- f.t
      # now in forecasting mode; loop over number of scenarios
      for (simulation_nr in sequence(scenarios)) {
        # print life sign
        if (as.integer(print_life_sign) > 0) if (simulation_nr %% as.integer(print_life_sign) == 0) print(paste("Iteration", as.integer(simulation_nr), "of", as.integer(scenarios), "(", round(100*simulation_nr/scenarios, 2), "%)"))
        # initialize f.t, yM, yR
        f.t <- f.t.original
        yM <- data$yM[nr_obs]
        yR <- data$yR[nr_obs, , ]
        forecasts$pi_emp_cum[simulation_nr, , ] <- diag(nr_out_ratings)
        forecasts$pi_cum[simulation_nr, , ] <- diag(nr_out_ratings)
        portfolio <- loanportfolio
        # loop over the forecasting horizon
        for (period_ahead in sequence(horizon)) {
          # compute the next probabilities
          tmp <- filter_one_step(parameter, f.t, yM, yR, zmatM[nr_obs + period_ahead - 1, ], zmatR[nr_obs + period_ahead - 1, ])
          f.t <- tmp$f.t
          # update the pi matrix with the updated f.t and the new covariates
          tmp$pi_matrix <- compute_pi_matrix(parameter, tmp$f.t, zmatM[nr_obs + period_ahead, ], zmatR[nr_obs + period_ahead, ])$pi_matrix
          # with these probabilities, simulate data
          simulation <- simulate_gas(parameter, tmp$f.t, tmp$pi_matrix, portfolio)
          yM <- simulation$yM + stress_scenario[period_ahead]
          yR <- simulation$yR
          portfolio <- colSums(yR[ , sequence(nr_out_ratings - 1)])
          # store data, factors, and probabilities
          forecasts$fM[simulation_nr, period_ahead] <- f.t[1]
          forecasts$yM[simulation_nr, period_ahead] <- yM
          forecasts$fR[simulation_nr, period_ahead] <- f.t[2]
          forecasts$yR[simulation_nr, period_ahead, , ] <- yR
          forecasts$pi[simulation_nr, period_ahead, , ] <- tmp$pi_matrix
          # update the expected cumulative frequency matrix
          pi_tmp[1:nr_in_ratings, ] <- tmp$pi_matrix
          forecasts$pi_cum[simulation_nr, , ] <- forecasts$pi_cum[simulation_nr, , ] %*% pi_tmp
          # update the empirical cumulative frequency matrix
          pi_tmp[1:nr_in_ratings, ] <- yR / matrix(rep(portfolio, nr_out_ratings), nrow = nr_in_ratings)
          forecasts$pi_emp_cum[simulation_nr, , ] <- forecasts$pi_emp_cum[simulation_nr, , ] %*% pi_tmp
        }
      }
    }
  }
  
  #################
  ## return
  #################
  if (horizon > 0) return(list(output = output, forecasts = forecasts)) else return(output)
}
