########################################################
########################################################
## Purpose: transform and get/set parameters
##
## INPUTS: 
## p: vector holding parameters if setpar!=TRUE, or list as below under 
##    output if setpar==TRUE.
## nr_in_ratings: number of input ratings (default = 2)
## nr_out_ratings: number of output ratings (default = 3) 
## nr_exogenousM: number of exogenous covariates for macro part (default = 0)
## nr_exogenousR: number of exogenous covariates for credit part (default = 0)
## setpar: boolean, if TRUE do the inverse transform
## usetransform: boolean, if TRUE use coded transform, otherwise use identity
##
## OUTPUTS: list()
##    $muM: intercept of macro
##    $omegaM: K x 1 matrix with coefficients of covariates in macro equation
##    $s2: variance of error term
##    $muR: I x (J-1) matrix of intercepts for the ordered logit,
##          with I nr of input ratings, and J number of output 
##          rating categories
##    $gammaR: I x 1 matrix of coefficients for macro factor
##    $deltaR: I x 1 matrix of coefficients for frailty factor
##    $omegaR: K x 1 matrix with coefficients of covariates in ratings equation
get_parameters <- function(p, nr_in_ratings = 2, nr_out_ratings = 3, 
                           nr_exogenousM = 0, nr_exogenousR = 0,
                           setpar = FALSE, usetransform = TRUE,
                           exog_names = NULL) {
  
  ############################
  ############################
  ## GET the parameters from 
  ## vector into a list()
  ############################
  ############################
  if (!setpar) {
    ## initialize
    i1 <- 1
    parameter <- list()
    
    ## get gas alpha parameters; exponential trafo
    parameter$alpha <- c(0,0)
    if (GLOBAL_WITH_MACRO_FRAILTY) {
      parameter$alpha[1] <- ifelse(usetransform, exp(p[i1]), p[i1])
      i1 <- i1 + 1
    }
    if (GLOBAL_WITH_CREDIT_FRAILTY) {
      parameter$alpha[2] <- ifelse(usetransform, exp(p[i1]), p[i1])
      i1 <- i1 + 1
      }
    
    ## get gas beta parameters; logit trafo
    parameter$beta <- c(0,0)
    if (GLOBAL_WITH_MACRO_FRAILTY) {
      if (GLOBAL_EWMA_MACRO) parameter$beta[1] <- 1 else {
        parameter$beta[1] <- ifelse(usetransform, 1 / (1 + exp(-p[i1])), p[i1])
        i1 <- i1 + 1
      }
    }
    if (GLOBAL_WITH_CREDIT_FRAILTY) {
      if (GLOBAL_EWMA_CREDIT) parameter$beta[2] <- 1 else {
        parameter$beta[2] <- ifelse(usetransform, 1 / (1 + exp(-p[i1])), p[i1])
        i1 <- i1 + 1
      }
    }
    
    ## get variance macro error; exp trafo
    parameter$s2 <- ifelse(usetransform, exp(p[i1]), p[i1])
    i1 <- i1 + 1
    
    ## get macro conditional mean
    parameter$muM <- p[i1]
    i1 <- i1 + 1
    
    ## get ordered logit intercepts; trafo to ensure decreasing
    parameter$muR <- matrix(p[i1 - 1 + sequence(nr_in_ratings * (nr_out_ratings - 1))], nrow = nr_in_ratings, ncol = nr_out_ratings - 1)
    if (usetransform) for (i2 in sequence(nr_out_ratings - 2)) parameter$muR[ , i2 + 1] <- parameter$muR[ , i2] - exp(parameter$muR[ , i2 + 1])
    i1 <- i1 + nr_in_ratings * (nr_out_ratings - 1)
    
    ## get macro factor loadings in credit
    if (GLOBAL_WITH_MACRO_FRAILTY_IN_CREDITS) {
      parameter$gammaR <- matrix(p[i1 - 1 + sequence(nr_in_ratings)], nrow = nr_in_ratings)
      i1 <- i1 + nr_in_ratings
    } else parameter$gammaR <- matrix(0, nrow = nr_in_ratings)
    
    ## get credit factor loadings in credit part; last element normalized to 1
    if (GLOBAL_WITH_CREDIT_FRAILTY) {
      if (nr_in_ratings > 1) {
        parameter$deltaR <- matrix(c(p[i1 -1 + sequence(nr_in_ratings-1)], 1), nrow = nr_in_ratings)
        i1 <- i1 + nr_in_ratings - 1
      } else parameter$deltaR <- 1
    } else parameter$deltaR <- matrix(0, nrow = nr_in_ratings)
    
    ## get coefficients for covariates
    if (GLOBAL_WITH_MACRO_COVARIATES) {
      parameter$omegaM <- matrix(p[i1 - 1 + sequence(nr_exogenousM)], nrow = nr_exogenousM)
      i1 <- i1 + nr_exogenousM
    } else parameter$omegaM <- 0
    if (GLOBAL_WITH_CREDIT_COVARIATES) {
      parameter$omegaR <- matrix(p[i1 - 1 + sequence(nr_in_ratings * nr_exogenousR)], ncol = nr_in_ratings, nrow = nr_exogenousR)
      i1 <- i1 + nr_in_ratings * nr_exogenousR
    } else parameter$omegaR <- 0
    
    
  ############################
  ############################
  ## SET the parameters from 
  ## list() into a vector
  ############################
  ############################
  } else {
    ## initialize
    parameter <- NULL
    parameternames <- NULL
    if (is.null(exog_names)) { 
      exog_names <- paste(sequence(nr_exogenousR), sep = "", collapse = "")
    }
    
    ## get gas alpha parameters; exponential trafo
    if (GLOBAL_WITH_MACRO_FRAILTY) {
      parameter <- c(parameter, ifelse(usetransform ,log(p$alpha[1]), p$alpha[1]))
      parameternames <- c(parameternames, ifelse(usetransform, "logalpha.M", "alpha.M"))
    }
    if (GLOBAL_WITH_CREDIT_FRAILTY) {
      parameter <- c(parameter, ifelse(usetransform ,log(p$alpha[2]), p$alpha[2]))
      parameternames <- c(parameternames, ifelse(usetransform, "logalpha.R", "alpha.R"))
    }

    ## get gas beta parameters; logit trafo
    if (GLOBAL_WITH_MACRO_FRAILTY & !GLOBAL_EWMA_MACRO) {
      parameter <- c(parameter, ifelse(usetransform, log(p$beta[1]/(1-p$beta[1])), p$beta[1]))
      parameternames <- c(parameternames, ifelse(usetransform, "logitbeta.M", "beta.M"))
    }
    if (GLOBAL_WITH_CREDIT_FRAILTY & !GLOBAL_EWMA_MACRO) {
      parameter <- c(parameter, ifelse(usetransform, log(p$beta[2]/(1-p$beta[2])), p$beta[2]))
      parameternames <- c(parameternames, ifelse(usetransform, "logitbeta.R", "beta.R"))
    }

    ## get variance macro error; exp trafo
    parameter <- c(parameter, ifelse(usetransform, log(p$s2), p$s2))
    parameternames <- c(parameternames, ifelse(usetransform, "logs2", "s2"))
    
    ## get macro conditional mean
    parameter <- c(parameter, p$muM)
    parameternames <- c(parameternames, "mu.M")
    
    ## get ordered logit intercepts; trafo to ensure decreasing
    muR <- p$muR
    if (usetransform) for (i2 in sequence(nr_out_ratings - 2)) muR[ , i2 + 1] <- log(p$muR[ , i2] - p$muR[ , i2 + 1])
    parameter <- c(parameter, c(muR))
    parameternames <- c(parameternames, 
                        paste("muR", rep(1:nr_in_ratings, nr_out_ratings - 1), 
                              kronecker(1:(nr_out_ratings - 1), rep(1, nr_in_ratings)), 
                              sep = "."))
    

    ## get macro factor loadings in credit
    if (GLOBAL_WITH_MACRO_FRAILTY_IN_CREDITS) {
      parameter <- c(parameter, p$gammaR)
      parameternames <- c(parameternames, paste0("gamma.R", 1:nr_in_ratings))
    }
    ## get credit factor loadings in credit part; last element normalized to 1
    if (GLOBAL_WITH_CREDIT_FRAILTY & (nr_in_ratings > 1)) {
      parameter <- c(parameter, p$deltaR[sequence(nr_in_ratings - 1)])
      parameternames <- c(parameternames, paste0("delta.R.", 1:(nr_in_ratings - 1)))
    }
    ## get coefficients for covariates
    if (GLOBAL_WITH_MACRO_COVARIATES) {
      parameter <- c(parameter, p$omegaM)
      parameternames <- c(parameternames, paste0("omega.M.", 1:nr_in_ratings))
    }
    if (GLOBAL_WITH_CREDIT_COVARIATES) {
      parameter <- c(parameter, c(p$omegaR))
      parameternames <- c(parameternames, 
                          paste("omegaR.r", 
                                kronecker(1:nr_in_ratings, rep(1, nr_exogenousR)),
                                ".", rep(exog_names, nr_in_ratings),
                                sep = ""))
    }
    names(parameter) <- parameternames
  }
    
  ## return parameter list 
  return(parameter)
}
